from .. import fields, manager, model
from ..interaction import GenericInteraction
from .note import NotesField
from .pipeline import Status
from .tag import TagsField


class _StatusesField(fields._BaseField):
    _path = []

    def __init__(self):
        super().__init__("status_id", blank=True)

    def on_get_instance(self, instance, status_id):
        for status in instance.pipeline.statuses:
            if status.id == status_id:
                return status

    def on_set_instance(self, instance, value):
        if isinstance(value, str):
            return Status.get_for(instance.pipeline).get(query=value).id
        if isinstance(value, Status):
            return value.id
        return value


class LeadsInteraction(GenericInteraction):
    path = "leads"


class Lead(model.Model):
    name = fields._Field("name")
    responsible_user = fields._Link("responsible_user_id", "User")
    group = fields._UnEditableField("group_id")
    price = fields._Field("price")
    created_by = fields._Link("created_by", "User")
    updated_by = fields._Link("updated_by", "User")
    created_at = fields._DateTimeField("created_at")
    updated_at = fields._DateTimeField("updated_at")
    closest_task_at = fields._DateTimeField("closest_task_at", blank=True)
    account_id = fields._UnEditableField("account_id")
    is_deleted = fields._UnEditableField("is_deleted")
    score = fields._UnEditableField("score", blank=True)

    loss_reason_id = fields._Field("loss_reason_id", blank=True)
    status = _StatusesField()
    pipeline = fields._Link("pipeline_id", "Pipeline")

    loss_reason = fields._Field("loss_reason", path=["_embedded"])
    contacts = fields._EmbeddedLinkListField("contacts", model="Contact")

    tags = TagsField()
    notes = NotesField()

    objects = manager.Manager(LeadsInteraction())



from .. import fields, manager, model
from ..interaction import GenericInteraction


class UsersInteraction(GenericInteraction):
    path = "users"


class User(model.Model):
    name = fields._Field("name")
    email = fields._Field("email")
    language = fields._Field("lang")
    mail_access = fields._Field("mail_access", path=["rights"])
    catalog_access = fields._Field("catalog_access", path=["rights"])
    status_rights = fields._Field("status_rights", path=["rights"], blank=True)
    is_admin = fields._Field("is_admin", path=["rights"])
    is_free = fields._Field("is_free", path=["rights"])
    is_active = fields._Field("is_active", path=["rights"])

    objects = manager.Manager(interaction=UsersInteraction())



from .. import fields, manager, model
from ..interaction import GenericInteraction


class StatusesInteraction(GenericInteraction):
    path = "leads/pipelines/{pipeline_id}/statuses"
    field = "statuses"

    def __init__(self, pipeline_id, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._pipeline_id = pipeline_id

    def _get_path(self):
        return self.path.format(pipeline_id=self._pipeline_id)

    def get_all(self, include=None, query=None, filters=(), order=None):
        for data in self._all(self._get_path(), include=include, query=None, filters=filters, order=order):
            for status_data in data[self._get_field()]:
                if not query or status_data["name"].lower() == query.lower():
                    yield status_data


class Status(model.Model):
    name = fields._Field("name")
    sort = fields._Field("sort")
    is_editable = fields._Field("is_editable")
    color = fields._Field("color")
    type = fields._Field("type")

    @classmethod
    def get_for(cls, pipeline):
        if isinstance(pipeline, Pipeline):
            pipeline = pipeline.id
        return manager.Manager(interaction=StatusesInteraction(pipeline_id=pipeline), model=cls)


class _StatusField(fields._Field):
    _path = ["_embedded"]

    def __init__(self):
        super().__init__(name="statuses")

    def on_get(self, data):
        return [Status(data=item) for item in data]


class PipelinesInteraction(GenericInteraction):
    path = "leads/pipelines"
    field = "pipelines"


class Pipeline(model.Model):
    name = fields._Field("name")
    sort = fields._Field("sort")
    is_main = fields._Field("is_main")
    is_unsorted_on = fields._Field("is_unsorted_on")
    is_archive = fields._Field("is_archive")
    account_id = fields._Field("account_id")

    statuses = _StatusField()

    objects = manager.Manager(PipelinesInteraction())



import pytest

from amocrm.v2 import exceptions
from amocrm.v2.entity.user import User

from .data.users import DETAIL_INFO


def test_get(response_mock):
    response_mock.add("GET", "https://test.amocrm.ru/api/v4/users/3", match_querystring=False, json=DETAIL_INFO)
    user = User.objects.get(3)

    assert user.id == 3
    assert user.name == "Алексей Поимцев"
    assert user.email == "test@example.com"
    assert user.language == "ru"
    assert not user.is_admin
    assert user.is_active



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

__all__ = ['settings']


class Settings(object):

    def __init__(self):
        self.user_login, self.user_hash, self.domain = None, None, None
        self.responsible_user = None

    def set(self, user_login, user_hash, domain, responsible_user=None):
        self.user_login, self.user_hash, self.domain = user_login, user_hash, domain
        if responsible_user is not None:
            self.responsible_user = responsible_user

    def get(self):
        return {
            'user_login': self.user_login,
            'user_hash': self.user_hash,
            'domain': self.domain,
            'responsible_user': self.responsible_user
        }

settings = Settings()



import jwt
import pytest

import responses
from amocrm.v2.tokens import TokensStorage, default_token_manager


class FakeStorage(TokensStorage):
    def get_access_token(self):
        return jwt.encode({}, "tests").decode()


@pytest.fixture(autouse=True)
def _token():
    default_token_manager(client_id="", client_secret="", subdomain="test", redirect_url="", storage=FakeStorage())


@pytest.fixture(name="response_mock")
def _mock():
    with responses.RequestsMock() as _responses:
        yield _responses



from .. import fields, manager, model
from ..interaction import GenericInteraction
from .note import NotesField
from .tag import TagsField


class ContactsInteraction(GenericInteraction):
    path = "contacts"


class Contact(model.Model):

    name = fields._Field("name")
    first_name = fields._Field("first_name")
    last_name = fields._Field("last_name")
    responsible_user = fields._Link("responsible_user_id", "User")
    created_by = fields._Link("created_by", "User")
    updated_by = fields._Link("updated_by", "User")
    created_at = fields._DateTimeField("created_at")
    updated_at = fields._DateTimeField("updated_at")
    closest_task_at = fields._DateTimeField("closest_task_at", blank=True)
    account_id = fields._UnEditableField("account_id")
    tags = TagsField()

    company = fields._EmbeddedLinkField("companies", "Company")
    leads = fields._EmbeddedLinkListField("leads", "Lead")

    notes = NotesField()

    objects = manager.Manager(ContactsInteraction())



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
import json
from abc import *
from collections import defaultdict
from time import time, sleep
from copy import copy
import logging
import six

import requests

from .settings import settings
from .utils import lazy_dict_property, lazy_property, User
from .exceptions import *

logger = logging.getLogger('amocrm')

__all__ = []

_AMO_LOGIN_PATH = '/private/api/auth.php?type=json'
_REQUEST_PARAMS = {
    'headers': {'User-Agent': 'Amocrm API module. Python powered'},
    'timeout': 10,
}
_G, _P = 'get', 'post'

_tree = lambda: defaultdict(_tree)
_session = requests.Session()


class _BaseAmoManager(six.with_metaclass(ABCMeta)):
    format_ = 'json'

    _object_type = None
    _base_path = '/private/api/v2/%(format)s%(name)s%(path)s'
    _methods = {
        'account_info': {
            'path': 'current',
            'result': ['account'],
            'name': 'accounts',
        },
        'list': {
            'path': 'list',
            'result': True,
        },
        'add': {
            'path': 'set',
            'method': _P,
            'result': ['add', 0, 'id'],
            'container': ['add'],
        },
        'links': {
            'path': 'links',
            'name': 'contacts',
            'result': ['links'],
        },
        'linkslist': {
            'path': 'list',
            'name': 'links',
            'result': ['links'],
        },
        'update': {
            'path': 'set',
            'method': _P,
            'container': ['update'],
            'result': ['update', 0, 'id'],
            'timestamp': True,
        },
    }
    _amo_model_class = None
    _main_field = None
    _container_name = None
    ObjectNotFound = ObjectNotFound

    def __init__(self, user_login=None, user_hash=None,
                 domain=None, responsible_user=None):
        self._session = _session
        if user_login is not None:
            settings.set(user_login, user_hash, domain,
                         responsible_user)

    def auth(self):
        try:
            self._make_request(_AMO_LOGIN_PATH, _P, data=self._login_data, headers = {'content-type': 'application/json'})
        except AmoResponseException as e:
            raise AmoAuthException(e.resp)

    @property
    def is_auth(self):
        return 'session_id' in self._session.cookies

    @property
    def _domain(self):
        return settings.domain

    @property
    def _login_data(self):
        if settings.user_login:
            return {'USER_LOGIN': settings.user_login,
                    'USER_HASH': settings.user_hash}

    @lazy_property
    def _responsible_user(self):
        return settings.responsible_user or settings.user_login

    @classmethod
    @abstractproperty
    def name(cls):
        pass

    @property
    def container_name(self):
        return self._container_name or self.name

    def _convert_to_obj(self, result):
        if result:
            if not self._amo_model_class:
                return result
            if isinstance(result, (tuple, list)):
                return (self._amo_model_class(obj, _loaded=True) for obj in result)
            return self._amo_model_class(result, _loaded=True)

    def create(self, **kwargs):
        return self.add(**kwargs)

    @lazy_dict_property
    def _custom_fields(self):
        return self.get_custom_fields(to=self.name)

    @lazy_dict_property
    def account_info(self):
        return self.get_account_info()

    @lazy_property
    def user(self):
        user = User.get_one(self.users, self._responsible_user)
        if user is None:
            raise Exception('Can not get responsible user id')
        return user

    @lazy_property
    def leads_statuses(self):
        return {item['name']: item for item in self.account_info.get('leads_statuses')}

    @lazy_property
    def all_leads_statuses(self):
        main = {item['id']: item for item in self.account_info.get('leads_statuses')}
        for pipeline in self.pipelines.values():
            main.update(pipeline['statuses'])
        return main

    @lazy_property
    def note_types(self):
        return {item['code']: item for item in self.account_info.get('note_types')}

    @lazy_property
    def task_types(self):
        return {item['name']: item for item in self.account_info.get('task_types')}

    @lazy_property
    def users(self):
        return [User(item) for item in self.account_info.get('users')]

    @lazy_property
    def pipelines(self):
        return {item['name']: item for item in self.account_info.get('pipelines').values()}

    def _make_request(self, path, method, data, headers=None):
        headers = headers or {}

        _req_params = copy(_REQUEST_PARAMS)
        headers.update(_req_params.pop('headers'))

        method = method.lower()
        params = {}
        if method != _P:
            params.update(data)
            data = None
            for name, value in params.items():
                if isinstance(value, (list, tuple)):
                    values = value
                    del params[name]
                    for i, value in enumerate(values):
                        key = '{0}[{1}]'.format(name, i)
                        params[key] = value
                elif isinstance(value, dict):
                    values = value
                    del params[name]
                    for item_key, item_value in values.items():
                        key = '{0}[{1}][{2}]'.format(name, 0, item_key)
                        params[key] = item_value

        logger.info('%s - Sending %s request to %s' % (self.__class__.name,
                                                       method, path))
        logger.debug('Data: %s \n Params: %s' % (data, params))
        resp = self._session.request(method, self._url(path), data=json.dumps(data), params=params,
                                     headers=headers, **_req_params)
        logger.debug('Url: %s', resp.url)
        if not resp.ok:
            logger.error('Something went wrong')
            if resp.status_code in (401, 403):
                raise AmoAuthException(resp)
            raise AmoResponseException(resp)
        try:
            return resp.json()
        except ValueError:
            return resp.content

    def _create_container(self, container, data):
        name = self.container_name
        container = ['request', name] + container
        _container = _ = _tree()
        for i, elem in enumerate(container):
            if i + 1 == len(container):
                _[elem] = data
                continue
            _ = _[elem]
        return _container

    def _modify_response(self, response, result):
        if isinstance(result, (list, tuple)):
            result = ['response', self.container_name] + result
            for key in result:
                try:
                    response = response[key]
                except (TypeError, KeyError, IndexError):
                    pass
        elif result:
            try:
                response = response['response'][self.container_name]
            except (TypeError, KeyError):
                pass
        return response

    def _request(self, method, data=None, headers=None, modified_since=None):
        if not self.is_auth:
            self.auth()
        path = self._get_path(method)
        _method = self._methods[method]
        method_type = _method.get('method', _G)
        timestamp, container, result = _method.get('timestamp'), _method.get('container'), _method.get('result')

        if modified_since:
            headers = {'if-modified-since': six.text_type(modified_since)}

        if timestamp:
            _time = timestamp if isinstance(timestamp, str) else 'last_modified'
            data[0].setdefault(_time, int(time()))
        if container is not None:
            data = self._create_container(container, data)

        retry = 3
        while True:
            try:
                response = self._make_request(path=path, method=method_type, data=data, headers=headers)
            except AmoAuthException:
                logger.debug('Got auth error... reauth and retry again')
                self._session = requests.Session()
                self.auth()
                retry -= 1
                if retry == 0:
                    raise
                continue
            except (AmoResponseException, requests.RequestException):
                logger.debug('Got error... retry again')
                sleep(5)
                retry -= 1
                if retry == 0:
                    raise
                continue

            return self._modify_response(response, result)

    def _get_path(self, method_name):
        name = self._methods[method_name].get('name', self.name)
        path = self._methods[method_name]['path']
        if not name.startswith('/'):
            name = '/' + name
        if not name.endswith('/') and not path.startswith('/'):
            name += '/'
        return self._base_path % {'path': path, 'format': self.format_, 'name': name}

    def _url(self, path):
        return 'https://%(domain)s.amocrm.ru%(path)s' % {'domain': self._domain, 'path': path}

    def get_custom_fields(self, to):
        custom_fields = self.account_info['custom_fields'].get(to, [])
        return {field['name']: field for field in custom_fields}

    def get_account_info(self):
        return self._request('account_info', data={})

    def _get_links(self, contacts=None, leads=None, limit=None, limit_offset=None):
        """ Low level api to get links between leads and contacts """
        data = {}
        if contacts:
            data['contacts_link'] = contacts
        elif leads:
            data['deals_link'] = leads
        if limit is not None:
            data['limit_rows'] = limit
        if limit_offset is not None:
            data['limit_offset'] = limit_offset
        return self._request('links', data=data)

    def _get_linkslist(self, from_type=None, from_ids=None, to_type=None, to_ids=None):
        data = {}

        if to_type:
            data['to'] = to_type
        if from_type:
            data['from'] = from_type
        if from_ids:
            data['from_id'] = from_ids
        if to_ids:
            data['to_id'] = to_ids

        data = {'links': data}
        return self._request('linkslist', data=data)

    def all(self, query=None, user=None, chunk=500, offset=0, **kwargs):
        limit = chunk
        while True:
            i = 0
            for item in self._all(limit=limit, limit_offset=offset, query=query, user=user, **kwargs):
                i += 1
                yield item
            if not i:
                break
            offset += i

    def _all(self, limit=500, limit_offset=None, query=None, user=None, **kwargs):
        request = query or {}
        if self._object_type:
            request.update({'type': self._object_type})
        if limit is not None:
            request['limit_rows'] = limit
        if limit_offset is not None:
            request['limit_offset'] = limit_offset
        if user:
            request['responsible_user_id'] = user.id if isinstance(user, User) else user
        response = self._request('list', data=request, **kwargs)
        return self._convert_to_obj(response) or []

    def get(self, id):
        results = self._all(limit=1, query={'id': id, 'type': self.container_name[:-1]})
        if not results:
            raise self.ObjectNotFound('%s with id %s not founded' % (self._amo_model_class, id))
        return list(results).pop()

    def search(self, query, limit=1, **kwargs):
        query = {'query': query}
        results = self._all(query=query, limit=limit, **kwargs)
        return list(results) if results else []

    def add(self, **kwargs):
        return self._request('add', data=[self._add_data(**kwargs)])

    def update(self, **kwargs):
        return self._request('update', data=[self._update_data(**kwargs)])

    def create_or_update(self, **kwargs):
        return self._create_or_update_data(**kwargs)

    def delete(self, id, name=None):
        name = name or self.container_name
        headers = {
            'Referer': self._url('/{0}/list/'.format(name)),
            'Accept': 'application/json, text/javascript, */*; q=0.01',
            'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8',
            'X-Requested-With': 'XMLHttpRequest',
        }
        return self._session.request(_P, self._url('/ajax/{0}/multiple/delete/'.format(name)),
                                     data={'ID[]': id, 'ACTION': 'DELETE'},
                                     headers=headers)

    @abstractmethod
    def _add_data(self, **kwargs):
        return kwargs

    @abstractmethod
    def _update_data(self, **kwargs):
        return kwargs

    @abstractmethod
    def _create_or_update_data(self, on_field=None, **data):
        query = data.get(on_field or self._main_field)
        obj = self.search(query) if query else {}
        if obj:
            obj = obj.pop()
            if any(obj.get(key) != data[key] for key in data.keys()):
                for key, value in data.items():
                    try:
                        setattr(obj, key, value)
                    except UneditableFieldError:
                        pass
                obj.save()
            return obj['id']
        else:
            return self.add(**data)


class _BlankMixin(object):
    def _add_data(self, **kwargs):
        return super(_BlankMixin, self)._add_data(**kwargs)

    def _update_data(self, **kwargs):
        return super(_BlankMixin, self)._update_data(**kwargs)

    def _create_or_update_data(self, **kwargs):
        return super(_BlankMixin, self)._create_or_update_data(**kwargs)


def _Helper(_class, name):
    class Mixin(object):
        def __init__(self, *args, **kwargs):
            super(Mixin, self).__init__(*args, **kwargs)
            setattr(self, name, _class(*args, **kwargs))
            setattr(getattr(self, name), '_account_info', self.account_info)

    return Mixin



DETAIL_INFO = {
    "id": 3,
    "name": "АО Рога и копыта",
    "responsible_user_id": 504141,
    "group_id": 0,
    "created_by": 504141,
    "updated_by": 504141,
    "created_at": 1582117331,
    "updated_at": 1586361223,
    "closest_task_at": None,
    "custom_fields_values": [
        {
            "field_id": 3,
            "field_name": "Телефон",
            "field_code": "PHONE",
            "field_type": "multitext",
            "values": [{"value": "123213", "enum_id": 1, "enum": "WORK"}],
        }
    ],
    "account_id": 10,
    "_links": {"self": {"href": "https://exmaple.amocrm.ru/api/v4/companies/1"}},
    "_embedded": {"tags": []},
}



from .entity import custom_field
from .entity.company import Company
from .entity.contact import Contact
from .entity.lead import Lead
from .entity.pipeline import Pipeline, Status
from .entity.tag import Tag
from .entity.user import User



# -*- coding: utf-8 -*-

__all__ = ['AmoAuthException', 'AmoResponseException', 'ObjectNotFound', 'UneditableFieldError']


class AmoApiException(Exception):
    pass


class ObjectNotFound(AmoApiException):
    pass


class AmoResponseException(AmoApiException):

    def __init__(self, resp):
        self.resp = resp
        self.msg = self.message = '%s, %s' % (resp.status_code, resp.content)


class AmoAuthException(AmoResponseException):
    pass


class UneditableFieldError(AmoApiException):
    pass


import pytest

from amocrm.v2.model import _get_container_by_path


@pytest.mark.parametrize(
    ("path", "data", "expect"),
    [
        (["name"], {"name": "test"}, {"name": "test"}),
        (["name"], {"name": "test", "value": "no"}, {"name": "test"}),
        (["name", "value"], {"name": {"value": "test"}}, {"name": {"value": "test"}}),
        (["name", "value"], {"name": {"value": "test", "second": "No"}}, {"name": {"value": "test"}}),
    ],
)
def test_container(path, data, expect):
    assert _get_container_by_path(path, data) == expect



# -*- coding: utf-8 -*-

import types

__all__ = ['lazy_dict_property', 'lazy_property', 'cached_property']


empty = type('empty', (), {})


class lazy_property(object):

    def __init__(self, calculate_function):
        self._calculate = calculate_function

    def __get__(self, obj, _=None):
        if obj is None:
            return self
        value = self._calculate(obj)
        if value is not None:
            setattr(obj, self._calculate.__name__, value)
        return value


class lazy_dict_property(object):

    def __init__(self, calculate_function):
        self.__delegate()
        self.fake = empty()
        self.fake._calculate = calculate_function

    def __get__(self, obj, _=None):
        self.fake._obj = obj
        return self.fake

    def __delegate(self):
        for method_name in dict.__dict__.keys():
            if hasattr(empty, method_name):
                continue
            meth = self.__dispatch(method_name)
            setattr(empty, method_name, meth)

    @staticmethod
    def __dispatch(name):
        def __wrapper(*args, **kwargs):
            this = args[0]
            value = this._calculate(this._obj)
            setattr(this._obj, this._calculate.__name__, value)
            return getattr(value, name)(*args[1:], **kwargs)
        return __wrapper


class cached_property(object):
    def __init__(self, func):
        self.__func = func
        self.__cache = None

    def __get__(self, obj, _=None):
        if obj is None:
            return self
        if self.__cache:
            return self.__cache
        value = self.__func(obj)
        if value is not None:
            self.__cache = value
        if isinstance(self.__cache, types.GeneratorType):
            self.__cache = list(self.__cache)
        return self.__cache


class User(object):
    def __init__(self, data):
        assert isinstance(data, dict)
        self.__data = data
        self.id = data['id']
        self.login, self.name = data.get('login'), data.get('name')
        self.phone = data.get('phone_number')

    def __repr__(self):
        return 'User(%s)' % self.name

    @classmethod
    def get_user(cls, name):
        return cls.get_one(cls.all(), [name])

    @classmethod
    def all(cls):
        from .api import AmoApi
        return AmoApi().users

    @staticmethod
    def get_one(array, values):
        filter_func = lambda _: _.login in values or _.name in values or _.id in values
        match = [user for user in array if filter_func(user)]
        return match.pop() if match else None


from datetime import datetime

from .. import fields, manager, model
from ..interaction import GenericInteraction

TEXT = "text"  # Текст
NUMERIC = "numeric"  # Число
CHECKBOX = "checkbox"  # Флаг
SELECT = "select"  # Список
MULTISELECT = "multiselect"  # Мультисписок
DATE = "date"  # Дата
URL = "url"  # Ссылка
TEXTAREA = "textarea"  # Текстовая область
RADIOBUTTON = "radiobutton"  # Переключатель
STREET_ADDRESS = "streetaddress"  # Короткий адрес
DATETIME = "date_time"  # Дата и время

MULTITEXT = "multitext"
# smart_address	Адрес
# birthday	День рождения
# legal_entity	Юр. лицо
# PRICE  = "price" # Цена
# CATEGORY = "category" # Категория
# ITEMS = "items"  # Предметы


class CustomFieldModel(model.Model):
    name = fields._Field("name")
    code = fields._Field("code")
    sort = fields._Field("sort")
    type = fields._Field("type")
    entity_type = fields._Field("entity_type")

    is_deletable = fields._Field("is_deletable", blank=True)
    is_api_only = fields._Field("is_api_only", blank=True)

    enums = fields._Field("enums", blank=True)

    @classmethod
    def get_for(cls, instance):
        return manager.Manager(
            GenericInteraction(path=f"{instance._path}/custom_fields", field="custom_fields",), model=CustomFieldModel,
        ).all()

    @classmethod
    def create_for(cls, instance, name, code=None, sort=None):
        return (
            manager.Manager(
                GenericInteraction(path=f"{instance._path}/custom_fields", field="custom_fields",),
                model=CustomFieldModel,
            )
            .create(name=name, code=code, sort=sort,)
            .code
        )


class BaseCustomField(fields._BaseField):
    _defaults_create_with = {}
    _real_code = None

    def __init__(self, name, code=None, real_code=None, auto_create=False, **kwargs):
        super().__init__("custom_fields_values", blank=True)
        self._name = name
        self._code = code
        self._create_with = kwargs
        self._real_code = self._real_code or real_code
        self._field_id = None
        self._auto_create = auto_create

    def _find(self, instance):
        for field in CustomFieldModel.get_for(instance):
            if field.name == self._name or (self._code and self._code == field.code):
                return field
        return None

    def _check(self, instance):
        if self._field_id:
            return
        field = self._find(instance)
        if field:
            self._real_code = field.code
            self._field_id = field.id
        # if self._real_code is None and self._auto_create:
        #     self._real_code = CustomFieldModel.create_for(
        #         instance=instance,
        #         name=self._name,
        #         code=self._code,
        #         **self._create_with,
        #         **self._defaults_create_with
        #     )
        #     return
        # raise Exception("Field does not exists")

    def on_get_instance(self, instance, data):
        if data is None:
            return
        field_data = self._get_raw_field(data)
        if field_data:
            return self.on_get(field_data)
        return self.on_get(None)

    def _get_raw_field(self, data):
        if data is None:
            return None
        for field in data:
            if field.get("field_name", "error") == self._name:
                return field
            if self._real_code and field.get("field_code") == self._real_code:
                return field
        return None

    def _create_raw_field(self):
        _data = {"field_id": self._field_id, "values": []}
        if self._real_code:
            _data["field_code"] = self._real_code
        if self._name:
            _data["field_name"] = self._name
        return _data

    def __set__(self, instance, value):
        if instance is None:
            return
        data = instance._data
        for _path in self._path:
            data = data.setdefault(_path, {self.name: []})
        _data = self._get_raw_field(data.get(self.name))
        if _data is None:
            self._check(instance)
            data.setdefault(self.name, []).append(self._create_raw_field())
            _data = self._get_raw_field(data.get(self.name))
        values = self.on_set_instance(_data["values"], value)

        _data["values"] = values
        self._notify_instance(instance)
        return values


class TextCustomField(BaseCustomField):
    _defaults_create_with = {"type": TEXT}

    def on_get(self, values):
        return values[0]["value"]

    def on_set(self, value):
        return [{"value": value}]


class TextAreaCustomField(TextCustomField):
    _defaults_create_with = {"type": TEXTAREA}


class UrlCustomField(TextCustomField):
    _defaults_create_with = {"type": URL}


class StreetAddressCustomField(TextCustomField):
    _defaults_create_with = {"type": STREET_ADDRESS}


class SelectCustomField(TextCustomField):
    _defaults_create_with = {"type": SELECT}

    def __init__(self, *args, enums=(), **kwargs):
        kwargs["enums"] = [{"value": enums[i], "sort": i + 1} for i in range(len(enums))]
        super().__init__(*args, **kwargs)


class RadioButtonCustomField(SelectCustomField):
    _defaults_create_with = {"type": RADIOBUTTON}


class MultiSelectCustomField(SelectCustomField):
    _defaults_create_with = {"type": MULTISELECT}

    def on_get(self, values):
        return [item["value"] for item in values]

    def on_set(self, values):
        return [{"value": value} for value in values]


class NumericCustomField(TextCustomField):
    _defaults_create_with = {"type": NUMERIC}

    def on_get(self, values):
        return float(values[0]["value"])


class CheckboxCustomField(TextCustomField):
    _defaults_create_with = {"type": CHECKBOX}

    def on_get(self, values):
        return bool(values[0]["value"])


class DateCustomField(TextCustomField):
    _defaults_create_with = {"type": DATE}

    def on_get(self, values):
        return datetime.fromtimestamp(values[0]["value"]).date()

    def on_set(self, value):
        if isinstance(value, datetime):
            value = value.timestamp()
        return super().on_set(value)


class DateTimeCustomField(DateCustomField):
    _defaults_create_with = {"type": DATETIME}

    def on_get(self, values):
        return datetime.fromtimestamp(values[0]["value"])


class ContactPhoneField(TextCustomField):
    _defaults_create_with = {"type": MULTITEXT}
    _real_code = "PHONE"

    def __init__(self, *args, enum_code="WORK", **kwargs):
        self._enum_code = enum_code
        super().__init__(*args, name=None, **kwargs)

    def on_get(self, values):
        for value in values:
            if value["enum_code"] == self._enum_code:
                return value["value"]

    def on_set_instance(self, values, value):
        for item in values:
            if item["enum_code"] == self._enum_code:
                item["value"] = value
                return values
        values.append({"enum_code": self._enum_code, "value": value})
        return values


class ContactEmailField(ContactPhoneField):
    _defaults_create_with = {"type": MULTITEXT}
    _real_code = "EMAIL"



from . import fields, manager, model
from .interaction import BaseInteraction


class AccountInteraction(BaseInteraction):
    def get(self, object_id=None, include=None):
        response, _ = self.request("get", "account", include=include)
        return response


class Account(model.Model):
    name = fields._Field("name")
    subdomain = fields._Field("subdomain")
    created_by = fields._Link("created_by", "User")
    updated_by = fields._Link("updated_by", "User")
    current_user = fields._Link("current_user_id", "User")
    created_at = fields._DateTimeField("created_at")
    updated_at = fields._DateTimeField("updated_at")
    country = fields._Field("country")
    customers_mode = fields._Field("customers_mode")
    is_unsorted_on = fields._Field("is_unsorted_on")
    is_loss_reason_enabled = fields._Field("is_loss_reason_enabled")
    is_helpbot_enabled = fields._Field("is_helpbot_enabled")
    is_technical_account = fields._Field("is_technical_account")
    contact_name_display_order = fields._Field("contact_name_display_order")
    version = fields._Field("version", is_embedded=True)

    users_groups = fields._Field("users_groups", path=["_embedded"])
    task_types = fields._Field("task_types", path=["_embedded"])
    datetime_settings = fields._Field("datetime_settings", path=["_embedded"])
    entity_names = fields._Field("entity_names", path=["_embedded"])


def get_account_info(*args, **kwargs):
    return manager.Manager(AccountInteraction(*args, **kwargs), model=Account).get()



import requests

from . import exceptions
from .tokens import default_token_manager

_session = requests.Session()


class BaseInteraction:
    _default_headers = {
        "User-Agent": "amocrm-py/v2",
    }

    def __init__(self, token_manager=default_token_manager, session=_session, headers=_default_headers):
        self._token_manager = token_manager
        self._session = session
        self._default_headers = headers

    def get_headers(self):
        headers = {}
        headers.update(self._default_headers)
        headers.update(self._get_auth_headers())
        return headers

    def _get_auth_headers(self):
        return {"Authorization": "Bearer " + self._token_manager.get_access_token()}

    def _get_url(self, path):
        return "https://{subdomain}.amocrm.ru/api/v4/{path}".format(subdomain=self._token_manager.subdomain, path=path)

    def _request(self, method, path, data=None, params=None, headers=None):
        headers = headers or {}
        headers.update(self.get_headers())
        response = self._session.request(method, url=self._get_url(path), json=data, params=params, headers=headers)
        if response.status_code == 204:
            return None, 204
        if response.status_code < 300 or response.status_code == 400:
            return response.json(), response.status_code
        if response.status_code == 401:
            raise exceptions.UnAuthorizedException()
        if response.status_code == 403:
            raise exceptions.PermissionsDenyException()
        if response.status_code == 402:
            raise ValueError("Тариф не позволяет включать покупателей")
        raise exceptions.AmoApiException("Wrong status {}".format(response.status_code))

    def request(self, method, path, data=None, params=None, headers=None, include=None):
        params = params or {}
        if include:
            params["with"] = ",".join(include)
        return self._request(method, path, data=data, params=params, headers=headers)

    def _list(self, path, page, include=None, limit=250, query=None, filters=(), order=None):
        assert order is None or len(order) == 1
        assert limit <= 250
        params = {
            "page": page,
            "limit": limit,
            "query": query,
        }
        if order:
            field, value = list(order.items())[0]
            params["order[{}]".format(field)] = value
        for _filter in filters:
            params.update(_filter._as_params())
        return self.request("get", path, params=params, include=include)

    def _all(self, path, include=None, query=None, filters=(), order=None, limit=250):
        page = 1
        while True:
            response, _ = self._list(
                path, page, include=include, query=query, filters=filters, order=order, limit=limit
            )
            if response is None:
                return
            yield response["_embedded"]
            if len(response["_embedded"]) != limit:
                return
            if "next" not in response["_links"]:
                return
            page += 1


class Filter:
    def __init__(self, name):
        self._name = name

    def _as_params(self):
        return {}


class SingleFilter(Filter):
    def __call__(self, value):
        self._value = value

    def _as_params(self):
        return {"filter[{}]".format(self._name): self._value}


class SingleListFilter(Filter):
    def __call__(self, value):
        self._value = value

    def _as_params(self):
        return {"filter[{}][]".format(self._name): self._value}


class MultiFilter(Filter):
    def __call__(self, values):
        self._values = values

    def _as_params(self):
        return {"filter[{}][0]".format(self._name): self._values}


class RangeFilter(Filter):
    def __call__(self, value_from, value_to):
        self._value_from = value_from
        self._value_to = value_to

    def _as_params(self):
        return {
            "filter[{}][from]".format(self._name): self._value_from,
            "filter[{}][to]".format(self._name): self._value_to,
        }


class GenericInteraction(BaseInteraction):
    path = ""
    field = None

    def __init__(self, *args, path=None, field=None, **kwargs):
        super().__init__(*args, **kwargs)
        if path is not None:
            self.path = path
        if field is not None:
            self.field = field

    def _get_field(self):
        return self.field or self.path

    def _get_path(self):
        return self.path

    def get_list(self, page, include=None, limit=250, query=None, filters=None, order=None):
        response, _ = self._list(
            self._get_path(), page, include=include, limit=limit, query=query, filters=filters, order=order
        )
        return response["_embedded"][self._get_field()]

    def get_all(self, include=None, query=None, filters=(), order=None):
        for data in self._all(self._get_path(), include=include, query=query, filters=filters, order=order):
            yield from data[self._get_field()]

    def get(self, object_id, include=None):
        path = "{}/{}".format(self._get_path(), object_id)
        response, status = self.request("get", path, include=include)
        if status == 204:
            raise exceptions.NotFound()
        return response

    def create(self, data):
        response, status = self.request("post", self._get_path(), data=[data])
        if status == 400:
            raise exceptions.ValidationError(response)
        return response["_embedded"][self._get_field()][0]

    def update(self, object_id, data):
        path = "{}/{}".format(self._get_path(), object_id)
        response, status = self.request("patch", path, data=data)
        if status == 400:
            raise exceptions.ValidationError(response)
        return response



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import json
import time
import copy
from datetime import datetime
import logging

import six
from pytz import utc, timezone

from . import fields
from .api import *


__all__ = ['BaseCompany', 'BaseContact', 'BaseLead',
           'CompanyTask', 'ContactTask', 'LeadTask',
           'CompanyNote', 'ContactNote', 'LeadNote', 'TaskNote']
KIEV = 'Europe/Kiev'
MOSCOW = 'Europe/Moscow'


logger = logging.getLogger(__name__)


class _ModelMeta(type):
    def __new__(mcs, name, bases, attrs):
        attrs.setdefault('_fields', {})
        for base in bases:
            for _name, _field in getattr(base, '_fields', {}).items():
                if _name not in attrs:
                    attrs[_name] = _field

        attrs['_fields'].update({name: instance for name, instance in attrs.items()
                                 if isinstance(instance, (fields._BaseField, fields.CustomField))})
        attrs['_required'] = [f.field for f in attrs['_fields'].values() if f.required]
        super_new = super(_ModelMeta, mcs).__new__(mcs, name, bases, attrs)
        _manager = getattr(super_new, 'objects', None)
        if _manager:
            _manager._amo_model_class = super(_ModelMeta, mcs).__new__(mcs, name, bases, attrs)
        return super_new


class _BaseModel(six.with_metaclass(_ModelMeta)):

    _ELEMENT_TYPES = {
        'contact': 1,
        'lead': 2,
        'company': 3,
        'task': 4,
    }

    id = fields._UneditableField('id', required=True)
    date_create = fields._StaticDateTimeField('date_create')
    last_modified = fields._StaticDateTimeField('last_modified')
    request = fields._Field('request_id')
    amo_user = rui = fields.Owner()
    amo_creator = fields.Owner('created_user_id')

    def __init__(self, data=None, **kwargs):
        self._data, self._init_data = {}, {}
        self._fields_data, self._changed_fields = {}, []
        self._loaded = bool(kwargs.pop('_loaded', False))
        if self._loaded:
            self._data = data or kwargs
        else:
            self._init_data = data or kwargs

            for name, field in self._fields.items():
                if isinstance(field, fields.ManyForeignField):
                    continue
                value = self._init_data.get(name, None)
                if value is None:
                    continue
                if isinstance(field, fields.ForeignField) and name in self._init_data:
                    main_field = field.object_type.objects._main_field
                    if isinstance(value, field.object_type):
                        setattr(self, name, value)
                    elif main_field in field.links.keys():
                        self._data[field.links[main_field]] = self._init_data[name]
                        self._changed_fields.append(field.field)
                else:
                    if isinstance(field, fields._UneditableField):
                        self._data[field.field] = value
                    else:
                        setattr(self, name, value)
                    self._changed_fields.append(field.field)

    def __getitem__(self, item):
        return self.__getattribute__(item)

    def __getattribute__(self, name):
        get_attr = super(_BaseModel, self).__getattribute__
        value = get_attr(name)
        if (value is None and not get_attr('_loaded') and
                    name != 'id' and
                    get_attr('_data').get('id', None) is not None and
                    name in get_attr('_fields') and
                    get_attr('_fields')[name].field not in get_attr('_changed_fields')):
            get_attr('_init')()
            value = get_attr(name)
        return value

    get = __getitem__

    def _init(self):
        if not self._loaded and self.id is not None:
            amo_data = self.objects.get(self.id)  # trying to get info from crm
            if amo_data:
                self.__init__(amo_data._data, _loaded=True)
                return True

    def _save_fk(self):
        for name, field in self._fields.items():
            if not isinstance(field, fields.ForeignField):
                continue
            main_field = field.object_type.objects._main_field
            if main_field is not None and main_field in field.links:
                value = getattr(getattr(self, name), main_field)
                self._data[field.links[main_field]] = value
            if getattr(field, 'auto', None):
                continue
            if (field.field in self._changed_fields or not self._loaded) \
                    and (name in self._data or name in self._init_data):
                if not getattr(self, name).id:  # save fk if it not exists
                    getattr(self, name).save()
                self._data[field.field] = getattr(self, name).id

    def _pre_save(self):
        if self.id and not self._data['id']:  # TODO: WTF ???
            self._data['id'] = self.id
        if self.date_create is None:
            self._data['date_create'] = int(time.time())
        current_last_modified = self._data.get('last_modified')
        now = int(time.time())
        if current_last_modified and current_last_modified > now:
            # Воркэраунд для ситуации, когда last_modified объекта на сервере больше, чем наше время
            logger.debug('last_modified workaround old last_modified=%s > now=%s', current_last_modified, now)
            self._data['last_modified'] = current_last_modified + 1
        else:
            self._data['last_modified'] = now
        for name, field in self._fields.items():
            if field.field in self._required:
                value = getattr(self, name)
                if not value:
                    continue
                if isinstance(field, fields._UneditableField):
                    self._data[field.field] = value
                else:
                    setattr(self, name, value)

        multi = [custom_field['name'] for custom_field in self.objects._custom_fields.values()
                 if custom_field['type_id'] == fields.MULTI_LIST_TYPE]
        if self._data.get(fields.CustomField._field, None):
            for field in multi:
                data = [cfield for cfield in self._data[fields.CustomField._field] if cfield.get('name', None) == field]
                data = data.pop() if data else None
                if data and data['values'] and isinstance(data['values'][0], dict):
                    data['values'] = [val['enum'] for val in data['values']]
            self._data[fields.CustomField._field] = [{key: value for key, value in cf.items() if key in ['id', 'values']}
                                                     for cf in self._data.get(fields.CustomField._field) if field in multi or cf['values'][0]['value']]

        if not self._loaded:
            for name, field in self._fields.items():
                if (isinstance(field, fields._BaseField) and field.required and getattr(self, name, None) is None)\
                        and not isinstance(field, fields._UneditableField):
                    raise ValueError('{} is required'.format(name))

    def save(self, update_if_exists=False):
        self._save_fk()
        self._pre_save()
        data = self._data
        if self.id is not None:
            method = self.objects.update
            if not self._changed_fields:
                return
            data = dict([(key, value) for key, value in data.items() if key in self._changed_fields or key in self._required])
        elif update_if_exists:
            method = self.objects.create_or_update
        else:
            method = self.objects.create
        result = method(**data)
        if result:
            self._data['id'] = result
        self._changed_fields = []
        self._fields_data = {}

    def _get_field_by_name(self, name):
        result = [v for k, v in self._fields.items() if v.field == name]
        return result.pop()

    def __str__(self):
        return '{self.__class__.__name__}({self.id})'.format(self=self)

    __repr__ = __str__

    def __hash__(self):
        return hash(self.id or json.dumps(self._data))

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if self.id == other.id:
                return True
            return False
        elif isinstance(other, six.string_types):
            param = getattr(self, 'name', None) or getattr(self, 'text', None)
            if param:
                return param == other

    def __ne__(self, other):
        return not self.__eq__(other)

    def get_copy(self):
        data = copy.deepcopy(self._data)
        data.pop('date_create')
        data.pop('id')
        return self.__class__(data=data, _loaded=True)

    @property
    def detail_url(self):
        if self.id:
            return self.objects._url('/{0}/detail/{1}'.format(self.objects.container_name, self.id))

    def delete(self):
        self.objects.delete(self.id)


class _AbstractNamedModel(_BaseModel):
    name = fields._Field('name', required=True)
    linked_leads = fields.ManyForeignField('linked_leads_id')
    tags = fields._TagsField('tags', 'name')


class BaseCompany(_AbstractNamedModel):
    contact_model = None
    type = fields._ConstantField('type', 'company')

    objects = CompanyManager()

    @property
    def notes(self):
        return CompanyNote.objects.all(query={'element_id': self.id})

    @property
    def tasks(self):
        return CompanyTask.objects.all(query={'element_id': self.id})

    def create_note(self, text, note_type='COMMON'):
        note = CompanyNote(company=self, type=note_type, text=text)
        note.save(update_if_exists=False)
        return note

    @property
    def contacts(self):
        return (
            (self.contact_model or BaseContact).objects.get(_id)
            for _id in set([
                item['to_id']
                for item in self.objects._get_linkslist(
                    from_type='company',
                    from_ids=[self.id],
                    to_type='contact',
                )
                if item['to'] == 'contacts'
            ])
        )


class BaseLead(_BaseModel):
    contact_model = None

    name = fields._Field('name', required=False)
    tags = fields._TagsField('tags', 'name')
    status = fields._StatusTypeField(required=True)
    pipeline = fields._TypeField('pipeline_id', choices='pipelines', required=False)
    company = fields.ForeignField(BaseCompany, 'linked_company_id',
                                  auto_created=True,
                                  links={'name': 'company_name'})
    budget = price = fields._Field('price')

    objects = LeadsManager()

    def create_task(self, text, complete_till, task_type='Follow-up'):
        task = LeadTask(lead=self, type=task_type, text=text, complete_till=complete_till)
        task.save(update_if_exists=False)
        return task

    @property
    def tasks(self):
        return LeadTask.objects.all(query={'element_id': self.id})

    def create_note(self, text, note_type='COMMON'):
        note = LeadNote(lead=self, type=note_type, text=text)
        note.save(update_if_exists=False)
        return note

    @property
    def notes(self):
        return LeadNote.objects.all(query={'element_id': self.id})

    @property
    def statuses(self):
        if self.pipeline and self.objects.pipelines[self.pipeline]:
            return {item.get('name', item.get('id')): item for item in self.objects.pipelines[self.pipeline]['statuses'].values()}
        return self.objects.leads_statuses

    @property
    def contacts(self):
        return ((self.contact_model or BaseContact).objects.get(_id) for _id in
                set([item['contact_id'] for item in self.objects._get_links(leads=[self.id])]))


class BaseContact(_AbstractNamedModel):
    leads_model = None
    leads = fields.ManyForeignField(BaseLead, 'linked_leads_id')
    type = fields._ConstantField('type', 'contact')
    company = fields.ForeignField(BaseCompany, 'linked_company_id',
                                  auto_created=True,
                                  links={'name': 'company_name'})
    created_user = fields._UneditableField('created_user')

    objects = ContactsManager()

    def create_task(self, text, complete_till, task_type='Follow-up'):
        task = ContactTask(contact=self, type=task_type, text=text, complete_till=complete_till)
        task.save(update_if_exists=False)
        return task

    @property
    def tasks(self):
        return ContactTask.objects.all(query={'element_id': self.id})

    def create_note(self, text, note_type='COMMON'):
        note = ContactNote(contact=self, type=note_type, text=text)
        note.save(update_if_exists=False)
        return note

    @property
    def notes(self):
        return ContactNote.objects.all(query={'element_id': self.id})

    @property
    def links(self):
        return (self.leads_model.objects.get(item['lead_id']) for item in self.objects._get_links(contacts=[self.id]))


class _AbstractTaskModel(_BaseModel):
    CALL = 'Звонок'
    MEETING = 'Встреча'
    LETTER = 'Письмо'
    FOLLOW = 'Follow-up'

    type = fields._TypeField('task_type', 'task_types', required=True)
    text = fields._Field('text', required=True)
    complete_till = fields._DateTimeField('complete_till', required=True)
    is_closed = fields._BooleanField('status', required=True)

    objects = TasksManager()

    @property
    def is_active(self):
        if self.is_closed or self.complete_till < datetime.utcnow():
            return False
        return True

    def get_task_time_tz(self, tz=KIEV):
        return utc.localize(self.complete_till).astimezone(timezone(tz))

    @property
    def is_meeting(self):
        return self.type == 'Встреча'

    @property
    def is_full_day(self):
        return self.complete_till.minute == 59

    def delete(self):
        self.objects.delete(self.id, name='todo')


class LeadTask(_AbstractTaskModel):
    lead = fields.ForeignField(BaseLead, 'element_id')
    _element_type = fields._ConstantField('element_type',
                                         _BaseModel._ELEMENT_TYPES['lead'])

    objects = TasksManager(object_type='lead')


class ContactTask(_AbstractTaskModel):
    contact = fields.ForeignField(BaseContact, 'element_id')
    _element_type = fields._ConstantField('element_type',
                                         _BaseModel._ELEMENT_TYPES['contact'])

    objects = TasksManager(object_type='contact')


class CompanyTask(_AbstractTaskModel):
    company = fields.ForeignField(BaseCompany, 'element_id')
    _element_type = fields._ConstantField('element_type',
                                         _BaseModel._ELEMENT_TYPES['company'])

    objects = TasksManager(object_type='company')


class _AbstractNoteModel(_BaseModel):
    type = fields._TypeField('note_type', 'note_types', required=True)
    text = fields._Field('text', required=True)

    objects = NotesManager()

    @property
    def properties(self):
        try:
            res = json.loads(self.text)
        except ValueError:
            res = None
        if not isinstance(res, dict):
            res = {'TEXT': self.text}
        return res


class LeadNote(_AbstractNoteModel):
    lead = fields.ForeignField(BaseLead, 'element_id')
    _element_type = fields._ConstantField('element_type',
                                          _BaseModel._ELEMENT_TYPES['lead'])

    objects = NotesManager(object_type='lead')


class ContactNote(_AbstractNoteModel):
    contact = fields.ForeignField(BaseContact, 'element_id')
    _element_type = fields._ConstantField('element_type',
                                          _BaseModel._ELEMENT_TYPES['contact'])

    objects = NotesManager(object_type='contact')


class CompanyNote(_AbstractNoteModel):
    company = fields.ForeignField(BaseCompany, 'element_id')
    _element_type = fields._ConstantField('element_type',
                                          _BaseModel._ELEMENT_TYPES['company'])

    objects = NotesManager(object_type='company')


class TaskNote(_AbstractNoteModel):
    task_id = fields._Field('element_id')
    _element_type = fields._ConstantField('element_type',
                                          _BaseModel._ELEMENT_TYPES['task'])

    objects = NotesManager(object_type='task')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
from .base import _BlankMixin, _BaseAmoManager, _Helper
from .utils import lazy_dict_property

import six

__all__ = ['AmoApi', 'NotesManager', 'ContactsManager', 'CompanyManager', 'LeadsManager', 'TasksManager']


class ContactsManager(_BlankMixin, _BaseAmoManager):
    name = 'contacts'
    _main_field = 'name'
    _object_type = 'contact'


class CompanyManager(_BlankMixin, _BaseAmoManager):
    name = 'company'
    _container_name = 'contacts'
    _object_type = name
    _main_field = 'name'

    @lazy_dict_property
    def _custom_fields(self):
        return self.get_custom_fields(to='companies')


class LeadsManager(_BlankMixin, _BaseAmoManager):
    name = 'leads'
    PRIMARY = 'Первичный контакт'
    CONVERSATION = 'Переговоры'
    DECIDE = 'Принимают решение'
    APPROVAL = 'Согласование договора'
    NOT_REALIZED = 'Закрыто и не реализовано'
    SUCCESSFULLY = 'Успешно реализовано'

    def all(self, query=None, status=None, **kwargs):
        query = query or {}
        if status:
            status = [status] if not isinstance(status, (list, tuple)) else status
            new_statuses = []
            for _status in status:
                if not ((isinstance(_status, six.string_types) and _status.isdigit()) or isinstance(_status, int)):
                    new_statuses.extend([k for k, v in self.all_leads_statuses.items() if v['name'] == _status])
                else:
                    new_statuses.append(_status)
            query['status'] = new_statuses
        return super(LeadsManager, self).all(query=query, **kwargs) or ()


class _ObjectIdMixin(object):
    def __init__(self, object_type=None, *args, **kwargs):
        super(_ObjectIdMixin, self).__init__(*args, **kwargs)
        self._object_type = object_type


class TasksManager(_ObjectIdMixin, _BlankMixin, _BaseAmoManager):
    name = 'tasks'
    _main_field = 'element_id'

    def search(self, *args, **kwargs):
        raise Exception('Amocrm havn\'t task search ability ')

    def _create_or_update_data(self, **data):
        return self.add(**data)


class NotesManager(_ObjectIdMixin, _BlankMixin, _BaseAmoManager):
    name = 'notes'
    _main_field = 'element_id'
    EXTERNAL_ATTACH = 'EXTERNAL_ATTACH'
    MAX_SYSTEM = 'MAX_SYSTEM'
    DEAL_STATUS_CHANGED = 'DEAL_STATUS_CHANGED'
    ATTACHEMENT = 'ATTACHEMENT'
    MAIL_MESSAGE = 'MAIL_MESSAGE'
    COMPANY_CREATED = 'COMPANY_CREATED'
    SMS_IN = 'SMS_IN'
    CALL_OUT = 'CALL_OUT'
    CONTACT_CREATED = 'CONTACT_CREATED'
    DEAL_CREATED = 'DEAL_CREATED'
    CALL = 'CALL'
    COMMON = 'COMMON'
    TASK_RESULT = 'TASK_RESULT'
    SMS_OUT = 'SMS_OUT'
    MAIL_MESSAGE_ATTACHMENT = 'MAIL_MESSAGE_ATTACHMENT'
    DROPBOX = 'DROPBOX'
    CALL_IN = 'CALL_IN'

    def search(self, *args, **kwargs):
        raise Exception('Amocrm havn\'t note search ability ')

    def _create_or_update_data(self, **data):
        return self.add(**data)


class AmoApi(_Helper(ContactsManager, 'contacts'), _Helper(CompanyManager, 'company'),
             _Helper(NotesManager, 'notes'), _Helper(LeadsManager, 'leads'), _Helper(TasksManager, 'tasks'),
             _BlankMixin, _BaseAmoManager):
    name = 'accounts'




from . import exceptions
from .interaction import BaseInteraction


class LinksInteraction(BaseInteraction):
    def link(self, for_entity, to_entity, main=False, metadata=None):
        return self._set("link", for_entity, to_entity, main=main, metadata=metadata)

    def unlink(self, for_entity, to_entity):
        return self._set("link", for_entity, to_entity)

    def _set(self, direction, for_entity, to_entity, main=False, metadata=None):
        path = "{}/{}/{}".format(for_entity._path, for_entity.id, direction)
        data = {"to_entity_id": to_entity.id, "to_entity_type": to_entity._path}
        if main:
            metadata = metadata or {}
            metadata["main"] = True
        data["metadata"] = metadata
        response, status = self.request("post", path, data=[data])
        if status == 400:
            raise exceptions.ValidationError(response)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import logging
from datetime import datetime
from calendar import timegm
from copy import deepcopy

import six
from .utils import User
from .exceptions import UneditableFieldError

__all__ = ['CustomField', u'EnumCustomField', 'ForeignField', 'ManyForeignField']

logger = logging.getLogger('amocrm')
CHOICE_TYPE = '4'
MULTI_LIST_TYPE = '5'


class _BaseField(object):
    def __init__(self, field=None, required=False):
        self.field = field
        self.required = required

    def __get__(self, instance, _=None):
        if instance is None:
            return self
        if instance._fields_data.get(self.field) is None:
            _data = instance._data.get(self.field)
            _data = self.on_get(_data, instance)
            instance._fields_data[self.field] = _data
        return instance._fields_data[self.field]

    def __set__(self, instance, value):
        if instance is None:
            return
        instance._fields_data[self.field] = None
        value = self.on_set(value, instance)
        if self.field not in instance._data or instance._data[self.field] != value:
            instance._data[self.field] = value
            instance._changed_fields.append(self.field)

    def on_set(self, value, instance):
        return value

    def on_get(self, data, instance):
        return data


class _Field(_BaseField):
    pass


class _UneditableField(_Field):
    def __set__(self, instance, value):
        raise UneditableFieldError


class _ConstantField(_UneditableField):
    def __init__(self, field=None, value=None):
        super(_ConstantField, self).__init__(field, True)
        self._data = value

    def on_get(self, data, instance):
        return self._data


class _DateTimeField(_Field):
    def on_get(self, data, instance):
        if data is not None:
            return datetime.utcfromtimestamp(float(data))

    def on_set(self, value, instance):
        if isinstance(value, datetime):
            return timegm(value.utctimetuple())


class _StaticDateTimeField(_UneditableField, _DateTimeField):
    pass


class _BooleanField(_Field):
    def on_get(self, data, instance):
        data = int(data) if str(data).isdigit() else data
        return bool(data)


class _BaseForeignField(_Field):
    def __init__(self, object_type=None, field=None, required=False):
        super(_BaseForeignField, self).__init__(field, required=required)
        self.object_type = object_type

    def __get__(self, instance, _=None):
        if instance._init():
            instance._fields_data[self.field] = None
        return super(_BaseForeignField, self).__get__(instance, _)


class ForeignField(_BaseForeignField):
    def __init__(self, object_type=None, field=None, auto_created=False,
                 links={}, required=False):
        super(ForeignField, self).__init__(object_type, field, required=required)
        self.auto, self.links = auto_created, deepcopy(links)
        self.links['id'] = field

    def on_get(self, data, instance):
        obj = self.object_type()
        obj._fields_data['id'] = instance._data.get(self.field)
        obj._data = {name: instance._data.get(value)
                     for name, value in self.links.items()}
        return obj

    def on_set(self, val, instance):
        if isinstance(instance._get_field_by_name(self.field), self.__class__):
            instance._fields_data[self.field] = val
            return val.id
        elif str(val).isdigit():
            return int(val)


class ManyForeignField(_BaseForeignField):
    def __init__(self, objects_type=None, field=None, key=None, required=False):
        super(ManyForeignField, self).__init__(field=field,
                                               object_type=objects_type,
                                               required=required)
        self.key = key

    def on_get(self, data, instance):
        if not data:
            return []
        if not instance._loaded:
            return data
        if isinstance(data, (list, tuple)):
            return [self.object_type().objects.get(item) for item in data]
        return self.object_type().objects.get(data)

    def on_set(self, value, instance):
        if value is None:
            instance._fields_data[self.field] = []
            return []
        if not isinstance(value, (list, tuple)):
            value = [value]
        instance._fields_data[self.field] = [item for item in value]
        return [int(item) if isinstance(item, (str, int)) else int(item.id) for item in value]


class _TagsField(_Field):
    def __init__(self, field=None, key=None, required=False):
        super(_TagsField, self).__init__(field, required=required)
        self.key = key

    def on_get(self, data, instance):
        if data:
            if isinstance(data, (list, tuple)):
                return [item['name'] for item in data]
            return data.replace(', ', ',').split(',')

    def on_set(self, value, *args):
        if value and value[0] and isinstance(value[0], dict):
            value = (item['name'] for item in value)
        return ', '.join(value)


class _TypeField(_Field):
    def __init__(self, field=None, choices=None, required=False):
        super(_TypeField, self).__init__(field, required=required)
        self.choices = choices

    def on_get(self, data, instance):
        if data:
            _data = [key for key, item in self.get_choices(instance).items()
                     if six.text_type(item['id']) == six.text_type(data)
                     or six.text_type(item.get('code', '')) == six.text_type(data)]
            data = _data.pop()
        return data

    def on_set(self, value, instance):
        return self.get_choices(instance)[value]['id']

    def get_choices(self, instance):
        return getattr(instance.objects, self.choices)


class _StatusTypeField(_TypeField):

    def __init__(self, **kwargs):
        super(_StatusTypeField, self).__init__('status_id', choices='leads_statuses', **kwargs)

    def get_choices(self, instance):
        ch = super(_StatusTypeField, self).get_choices(instance)
        ch.update(instance.statuses)
        return ch


class Owner(_Field):

    def __init__(self, field='responsible_user_id'):
        super(Owner, self).__init__(field, required=False)

    def on_get(self, data, instance):
        if data and str(data).isdigit():
            owners = [item for item in instance.objects.users if str(item.id) == str(data)]
            if owners:
                return owners.pop()

    def on_set(self, value, instance):
        if isinstance(value, six.string_types):
            return User.get_one(instance.objects.users, [value, ]).id
        elif value is not None:
            return value.id


class CustomField(object):
    _field = 'custom_fields'

    def __init__(self, custom_field, subtypes=False, required=False):
        self.field = '%s_%s' % (self._field, custom_field)
        self.custom_field, self.subtypes = custom_field, subtypes
        self.required = required

    def __get__(self, instance, _=None):
        if instance is None:
            return self
        if instance._fields_data.get(self.field) is None:
            _data = instance._data.get(self._field)

            if _data is None:
                return
            self._check_field(instance)
            custom_field_info = instance.objects._custom_fields[self.custom_field]
            self._id = _id = int(custom_field_info['id'])
            _data = list(_data.values()) if isinstance(_data, dict) else _data
            _data = [item['values'] for item in _data if item['id'] == _id]

            _data = _data[-1] if _data else None
            _data = list(_data.values()) if isinstance(_data, dict) else _data
            if custom_field_info['type_id'] == MULTI_LIST_TYPE and _data and not isinstance(_data[0], dict):
                _data = [{'value': item[1]} for item in custom_field_info['enums'].items()
                         if item[0] in _data]
            _data = [item['value'] for item in _data] if _data else None
            if custom_field_info['type_id'] != MULTI_LIST_TYPE and custom_field_info.get(u'multiple') != u'Y' and _data:
                _data = _data.pop()

            if _data in ['0', '1']:
                _data = bool(int(_data))

            instance._fields_data[self.field] = _data
        return instance._fields_data[self.field]

    def __set__(self, instance, value):
        if instance is None:
            return
        self._check_field(instance)
        if isinstance(value, bool):
            value = str(int(value))
        instance._fields_data[self.field] = None
        custom_field_info = instance.objects._custom_fields[self.custom_field]
        _id = custom_field_info['id']
        instance._data[self._field] = instance._data.get(self._field, None) or []
        fields = [item for item in instance._data[self._field] if item['id'] == _id]

        if isinstance(value, (list, tuple)):
            _elems = []
            for v in value:
                _elems.append({'value': v})
        else:
            _elems = [{'value': value}]

        if 'enums' in custom_field_info and custom_field_info['enums']:
            values = [item['value'] for item in _elems]
            _elems = [{'enum': item[0], 'value': item[1]} for item in custom_field_info['enums'].items()
                      if item[1] in values]
            if custom_field_info['type_id'] == MULTI_LIST_TYPE:
                _elems = [item[0] for item in custom_field_info['enums'].items()
                          if item[1] in values]

            if not _elems:
                raise ValueError('Invalid value for %s' % self.field)

        if not fields:
            full_data = {'id': _id, 'values': _elems, 'name': self.custom_field}
            if self.subtypes:
                full_data['values'] = [{'subtype': str(i),
                                        'value': val.strip()} for i, val in enumerate(value.split(';'), start=1)]
            instance._data[self._field].append(full_data)
        else:
            fields[0]['values'] = _elems
        if self._field not in instance._changed_fields:
            instance._changed_fields.append(self._field)

    def _check_field(self, instance):
        if self.custom_field not in instance.objects._custom_fields:
            raise ValueError('%s hasn\'t field "%s"' % (instance, self.custom_field))


class EnumCustomField(CustomField):

    def __init__(self, custom_field, enum):
        super(EnumCustomField, self).__init__(custom_field)
        self.field = self.field + enum
        self.enum = enum

    def __get__(self, instance, _=None):
        if instance is None:
            return self
        if instance._fields_data.get(self.field) is None:
            _data = instance._data.get(self._field)

            if _data is None:
                return
            if self.custom_field not in instance.objects._custom_fields:
                raise ValueError(u"%s have not custom field '%s'" % (instance.objects.name, self.custom_field))
            custom_field_info = instance.objects._custom_fields[self.custom_field]
            _id = int(custom_field_info['id'])
            _data = [item['values'] for item in _data if item['id'] == _id]

            _data = _data[-1] if _data else None
            enum = {enum_name: _id for _id, enum_name in custom_field_info.get('enums', {}).items()}.get(self.enum)
            if not enum:
                raise ValueError(u"%s have not custom field '%s' with enum %s" % (instance.objects.name,
                                                                                  self.custom_field,
                                                                                  self.enum))

            if _data is None:
                return
            _data = [item for item in _data if item.get('enum') == int(enum)]
            self._check_field(instance)
            instance._fields_data[self.field] = [item['value'] for item in _data] if _data else None
            if len(_data) == 1:
                instance._fields_data[self.field] = instance._fields_data[self.field].pop()

        return instance._fields_data[self.field]

    def __set__(self, instance, values):
        if instance is None:
            return
        if not isinstance(values, (list, tuple)):
            values = [values]
        self._check_field(instance)
        instance._fields_data[self.field] = None
        custom_field_info = instance.objects._custom_fields[self.custom_field]
        _id = int(custom_field_info['id'])
        field = [_field for _field in instance._data.setdefault(self._field, []) if _field['id'] == _id]
        enum = {enum: int(_id) for _id, enum in custom_field_info.get('enums', {}).items()}.get(self.enum)
        if field:
            _tmp_vals=[]
            for v in field[0]['values']:
                if v['enum'] != enum:
                    _tmp_vals += [v]

            _tmp_vals += [{'value': value, 'enum': enum} for value in values]
            field[0]['values']=_tmp_vals
        else:
            full_data = {'id': _id, 'values': [{'value': value, 'enum': enum} for value in values]}
            instance._data[self._field].append(full_data)
        if self._field not in instance._changed_fields:
            instance._changed_fields.append(self._field)

    def _check_field(self, instance):
        super(EnumCustomField, self)._check_field(instance)
        custom_field_info = instance.objects._custom_fields[self.custom_field]
        if {enum: _id for _id, enum in custom_field_info.get('enums', {}).items()}.get(self.enum) is None:
            raise Exception('There is no enum "%s" for a field "%s"' % (self.enum, self.custom_field))



DETAIL_INFO = {
    "id": 3,
    "name": "Алексей Поимцев",
    "email": "test@example.com",
    "lang": "ru",
    "rights": {
        "leads": {"view": "M", "edit": "M", "add": "D", "delete": "M", "export": "M"},
        "contacts": {"view": "M", "edit": "M", "add": "D", "delete": "M", "export": "M"},
        "companies": {"view": "M", "edit": "M", "add": "D", "delete": "M", "export": "M"},
        "tasks": {"edit": "A", "delete": "A"},
        "mail_access": False,
        "catalog_access": True,
        "status_rights": [
            {
                "entity_type": "leads",
                "pipeline_id": 3166396,
                "status_id": 142,
                "rights": {"view": "D", "edit": "D", "delete": "D", "export": "D"},
            },
            {
                "entity_type": "leads",
                "pipeline_id": 3166396,
                "status_id": 32311027,
                "rights": {"view": "D", "edit": "D", "delete": "D"},
            },
            {
                "entity_type": "leads",
                "pipeline_id": 3104455,
                "status_id": 31881115,
                "rights": {"view": "D", "edit": "D", "delete": "D"},
            },
        ],
        "is_admin": False,
        "is_free": False,
        "is_active": True,
        "group_id": None,
        "role_id": None,
    },
    "_links": {"self": {"href": "https://example.amocrm.ru/api/v4/users/185848"}},
}



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
import json
import os
import time
from functools import wraps

import six
if six.PY2:
    from urlparse import urlparse, parse_qsl
else:
    from urllib.parse import urlparse, parse_qsl

from responses import RequestsMock

DIR = os.path.normpath(os.path.dirname(__file__))
FILE_DIR = os.path.join(DIR, 'generated.json')


def check_auth(func):
    @wraps(func)
    def _call(self, obj, params):
        if self._check_auth(params):
            return func(self, obj, params)
        return json.dumps({'status': 'error', 'auth': False})

    return _call


class FakeApi(object):
    """docstring for FakeApi"""

    def __init__(self, data):
        self.login, self.hash = None, None
        self._data = data

    def _check_auth(self, params):
        return True

    def _auth(self, obj, params):
        r = {'auth': False}
        login = params.pop('USER_LOGIN', None)
        _hash = params.pop('USER_HASH', None)
        if login == self.login and _hash == self.hash:
            r['auth'] = True
        return json.dumps(r)

    @check_auth
    def _list(self, obj, params):
        if obj == 'company':
            obj = 'contacts'
        if params is None:
            return json.dumps({'status': 'error'})
        _id, query = params.get('id'), params.get('query')
        if 'type' in params:
            params.pop('type')
        if _id is not None:
            resp = [i for i in self._data[obj] if int(i['id']) == int(_id)]
        elif query:
            resp = [i for i in self._data[obj] if query in i.values()]
        else:
            resp = self._data[obj]
        if 'limit_rows' in params:
            resp = resp[int(params.get('limit_offset', 0)):int(params.get('limit_rows'))]
        return json.dumps({'response': {obj: resp}})

    @check_auth
    def _set(self, obj, params):
        if obj == 'company':
            obj = 'contacts'
        resp = {}
        try:
            params = params['request'][obj]
        except KeyError:
            pass
        if params:
            if 'update' in params:
                update = params['update']
                update = json.loads(update)[0] \
                    if isinstance(update, str) else update[0]
                update['last_modified'] = time.time()
                target_id = update['id']
                update_obj = next(i for i in self._data[obj]
                              if i['id'] == target_id)
                update_obj.update(update)
                resp = {'update': {'id': target_id}}
            elif 'add' in params:
                add = params['add']
                add = json.loads(add)[0] if isinstance(add, str) else add[0]
                add['last_modified'] = time.time()
                max_id = max(map(lambda x: int(x['id']), self._data[obj]))\
                            if self._data[obj] else 0
                _id = max_id + 1
                add['id'] = _id
                self._data[obj].append(add)
                resp = {'add': {'id': _id, 'request_id': 1}}
        return json.dumps({'response': {obj: resp}})

    @check_auth
    def _current(self, obj, params):
        return json.dumps({'response': {obj[:-1]: self._data.get(obj, self._data.get(obj[:-1]))}})


class AmoApiMock(RequestsMock):
    _objects = ('contacts', 'notes', 'company', 'tasks', 'leads')
    _base_url = 'https://%(domain)s.amocrm.ru/private/api/v2/%(format)s%(name)s%(path)s'
    _format = 'json'
    fake_api_cls = FakeApi

    API_DATA = {
        'contacts': [],
        'notes': [],
        'account': {
            'id': 1,
            'custom_fields': {
                'contacts': [
                    {
                        'id': 1,
                        'name': 'Телефон',

                        'type_id': '1',
                    },
                    {'id': 2, 'name': 'Email', 'type_id': '1'},
                ],
            },

            'users': [
                {'id': 99, 'name': 'test', 'login': 'test'}
            ],
            'task_types': [
                {'code': 'CALL', 'id': 1, 'name': 'Call'},
                {'code': 'MEETING', 'id': 2, 'name': 'Meeting'},
                {'code': 'LETTER', 'id': 3, 'name': 'Letter'},
            ],
            'leads_statuses': [
                {'color': '#99CCFF', 'sort': '10', 'editable': 'N', 'name': 'test1', 'id': '7495620'},
                {'color': '#FFFF99', 'sort': '20', 'editable': 'N', 'name': 'test2', 'id': '7495622'},
            ],
            'note_types': [
                {'code': 'DEAL_CREATED', 'editable': 'N', 'id': 1, 'name': ''},
                {'code': 'CONTACT_CREATED', 'editable': 'N', 'id': 2, 'name': ''},
                {'code': 'DEAL_STATUS_CHANGED', 'editable': 'N', 'id': 3, 'name': ''},
                {'code': 'COMMON', 'editable': 'Y', 'id': 4, 'name': ''},
            ],
        },
        'company': [],
        'leads': [],
        'tasks': [],
    }

    def reset(self):
        super(AmoApiMock, self).reset()
        self._faker = self.fake_api_cls(data=self.API_DATA)

    def _find_match(self, request):
        result = super(AmoApiMock, self)._find_match(request)
        if not result:
            result = self._get_response(request)
        return result

    def _get_response(self, request):
        url_parsed = urlparse(request.url)
        if url_parsed.query or request.body:
            url_qsl = dict(parse_qsl(url_parsed.query) + parse_qsl(request.body))
            try:
                body_data = json.loads(request.body)
            except ValueError:
                pass
            else:
                if body_data:
                    url_qsl.update(body_data)
        else:
            url_qsl = {}
        obj, method = url_parsed.path.split('/')[-2:]
        method = method.split('.')[0]
        body = getattr(self._faker, '_%s' % method)(obj, url_qsl)
        return {
            'body': bytearray(body, 'UTF-8'),
            'content_type': 'text/plain',
            'status': 200,
            'adding_headers': None,
            'stream': False
        }

    def set_login_params(self, login, _hash):
        self._faker.login = login
        self._faker.hash = _hash


amomock = AmoApiMock()



class BaseModuleException(Exception):
    """
    Base exception for all
    """


class AmoApiException(BaseModuleException):
    """
    Base exception for errors from the amocrm api side
    """


class UnAuthorizedException(AmoApiException):
    """
    Something wrong with access token or credations
    """


class PermissionsDenyException(AmoApiException):
    """
    No permissions
    """


class NotFound(AmoApiException):
    """
    Amocrm api return 404 or nothing
    """


class NoToken(BaseModuleException):
    """
    Raised if there are no access token in the storage
    Action - init token
    """


class NoDataException(BaseModuleException):
    """
    Exception raised when no data found for field
    Wrong mapping or field may be blank (blank=True)
    """


class ValidationError(BaseModuleException):
    """
    Exception for validation error response from amocrm
    """


class InitException(BaseModuleException):
    """
    Instance have no id, create instance first
    """



from .. import fields, manager, model
from ..interaction import GenericInteraction
from .tag import TagsField


class CustomersInteraction(GenericInteraction):
    path = "customers"

    def enable(self, mode: str = "segments"):
        return self.set(True, mode=mode)

    def disable(self):
        return self.set(False)

    def set(self, enable: bool, mode: str = "segments"):
        data, _ = self.request("patch", path="customers/mode", data={"is_enabled": enable, "mode": mode})
        return data


class Customer(model.Model):

    name = fields._Field("name")
    next_price = fields._Field("next_price")
    next_date = fields._DateTimeField("next_date")
    responsible_user = fields._Link("responsible_user_id", "User")
    periodicity = fields._Field("periodicity")
    created_by = fields._Link("created_by", "User")
    updated_by = fields._Link("updated_by", "User")
    created_at = fields._DateTimeField("created_at")
    updated_at = fields._DateTimeField("updated_at")
    closest_task_at = fields._DateTimeField("closest_task_at", blank=True)
    is_deleted = fields._Field("is_deleted")
    ltv = fields._Field("ltv")
    purchases_count = fields._Field("purchases_count")
    average_check = fields._Field("average_check")

    account_id = fields._UnEditableField("account_id")
    tags = TagsField()

    contacts = fields._EmbeddedLinkListField("contacts", "Contact")
    company = fields._EmbeddedLinkField("companies", "Company")

    objects = manager.Manager(CustomersInteraction())



from .. import fields, manager, model
from ..interaction import GenericInteraction

COMMON_TYPE = "common"  # Обычное примечание
CALL_IN_TYPE = "call_in"  # Входящий звонок
CALL_OUT_TYPE = "call_out"  # Исходящий звонок
SERVICE_MESSAGE_TYPE = "service_message"  # Системное сообщение
CASHIER_MESSAGE_TYPE = "message_cashier"  # Сообщение кассиру
INVOICE_PAID_TYPE = "invoice_paid"  # Оплата счета
GEOLOCATION_TYPE = "geolocation"  # Геолокация
SMS_IN_TYPE = "sms_in"  # Входящее сообщение
SMS_OUT_TYPE = "sms_out"  # Исходящее сообщение
# "extended_service_message	Системное примечание


class NotesInteraction(GenericInteraction):
    path = "{entity_type}/{entity_id}/notes"


class NotesField(fields._UnEditableField):
    def __init__(self):
        super().__init__(blank=True, is_embedded=False)

    def on_get_instance(self, instance, value):
        class Note(_Note):
            objects = manager.Manager(NotesInteraction(path=f"{instance._path}/{instance.id}/notes", field="notes"))

        return Note


class _AutoTypeField(fields._Field):
    def __init__(self, *args, note_type, **kwargs):
        self._auto_type = note_type
        super().__init__(*args, blank=True, path=["params"], **kwargs)

    def on_set_instance(self, instance, value):
        instance._data["note_type"] = self._auto_type
        return super().on_set_instance(instance, value)


class _Note(model.Model):
    note_type = fields._Field("note_type")
    created_by = fields._Link("created_by", "User")
    updated_by = fields._Link("updated_by", "User")
    created_at = fields._DateTimeField("created_at")
    updated_at = fields._DateTimeField("updated_at")
    responsible_user = fields._Link("responsible_user_id", "User")

    objects = manager.Manager(NotesInteraction())

    # Type fields

    text = _AutoTypeField("text", note_type=COMMON_TYPE)

    uniq = _AutoTypeField("uniq", note_type=CALL_IN_TYPE)
    duration = _AutoTypeField("duration", note_type=CALL_IN_TYPE)
    source = _AutoTypeField("source", note_type=CALL_IN_TYPE)
    link = _AutoTypeField("link", note_type=CALL_IN_TYPE)
    phone = _AutoTypeField("phone", note_type=CALL_IN_TYPE)

    address = _AutoTypeField("address", note_type=GEOLOCATION_TYPE)
    longitude = _AutoTypeField("longitude", note_type=GEOLOCATION_TYPE)
    latitude = _AutoTypeField("latitude", note_type=GEOLOCATION_TYPE)



import inspect

from . import fields
from .register import _RegisterMeta

_EMBEDDED_FIELDS = (fields._EmbeddedLinkListField, fields._EmbeddedLinkField)
_FIELDS_FOR_REPR = (fields._Field,) + _EMBEDDED_FIELDS


class Model(metaclass=_RegisterMeta):
    _init_data = {}
    id = fields._UnEditableField("id", blank=True)

    def __init__(self, data=None, **kwargs):
        self._data = data or {}
        self._data.update(self._init_data)
        self._updated_fields = set()
        for attr, value in kwargs.items():
            setattr(self, attr, value)

    @property
    def _path(self):
        return self._manager._interaction.path

    @classmethod
    def _get_embedded_fields(cls):
        return [
            field.name
            for _, field in inspect.getmembers(cls)
            if isinstance(field, fields._BaseField) and field.is_embedded
        ]

    def __repr__(self):
        fields = [
            "{} = {}".format(field.name, getattr(self, attr))
            for attr, field in inspect.getmembers(self.__class__)
            if isinstance(field, _FIELDS_FOR_REPR)
        ]
        return "{self.__class__.__name__}({fields})".format(self=self, fields=", ".join(fields))

    def save(self):
        if self.id:
            self.update()
        else:
            self.create()
        return self

    @property
    def _manager(self):
        return self.__class__.objects

    def create(self):
        self._data["id"] = self._manager.create(self._data).id

    def update(self):
        if self._updated_fields:
            self._manager.update(self.id, self._get_updated_data())
            self._updated_fields = set()

    def _get_updated_data(self):
        data = {}
        for field_path in self._updated_fields:
            data.update(_get_container_by_path(field_path, self._data))
        return data


def _get_container_by_path(path, data):
    container = {}
    if not path:
        return data
    container[path[0]] = _get_container_by_path(path[1:], data[path[0]])
    return container



from datetime import datetime

from . import exceptions
from .links import LinksInteraction
from .register import get_model_by_name


class _BaseField:
    _path = []

    def __init__(self, name=None, blank=False, path=None, cache=False, is_embedded=None):
        self.name = name
        self._blank = blank
        self._path = path if path is not None else self._path
        self._cache = cache
        self.__value = None
        self.__is_embedded = is_embedded

    def __repr__(self):
        return f"{self.__class__.__name__}(name={self.name}, blank={self._blank}, path={self._path})"

    def __hash__(self):
        return "{}:{}".format(self.name, ":".join(self._path))

    def __get__(self, instance, _=None):
        # TODO: cache for get and check set for difference
        if instance is None:
            return self
        data = instance._data
        for _path in self._path:
            data = data.get(_path, {})
        data = data.get(self.name)
        if data is None and not self._blank:
            raise exceptions.NoDataException(str(self))
        return self.on_get_instance(instance, data)

    def __set__(self, instance, value):
        if instance is None:
            return
        value = self.on_set_instance(instance, value)
        data = instance._data
        for _path in self._path:
            data = data.setdefault(_path, {})
        if self.name not in data or data[self.name] != value:
            data[self.name] = value
            self._notify_instance(instance)
        return value

    def _notify_instance(self, instance):
        instance._updated_fields.add((*self._path, self.name))

    def on_set_instance(self, instance, value):
        return self.on_set(value)

    def on_get_instance(self, instance, data):
        return self.on_get(data)

    def on_set(self, value):
        return value

    def on_get(self, data):
        return data

    @property
    def is_embedded(self):
        if self.__is_embedded is None:
            return bool(self._path and self._path[0] == "_embedded")
        return self.__is_embedded


class _Field(_BaseField):
    pass


class _UnEditableField(_Field):
    def __set__(self, instance, value):
        raise TypeError()


class _DateTimeField(_Field):
    def on_get(self, data):
        if data is not None:
            return datetime.utcfromtimestamp(float(data))

    def on_set(self, value):
        if isinstance(value, datetime):
            return value.timestamp()


class _ObjectField(_UnEditableField):
    def __init__(self, model, *args, many=False, **kwargs):
        self._model = model
        self._many = many
        super().__init__(*args, **kwargs)

    def on_get(self, data):
        if self._many:
            return [self._model.from_dict(**item) for item in data]
        return self._model.from_dict(**data)


class _Link(_BaseField):
    def __init__(self, name, model, links=LinksInteraction(), manager=None):
        super().__init__(name, blank=True)
        self.__model = model
        self._links = links
        self.__manager = manager

    @property
    def _model(self):
        if isinstance(self.__model, str):
            self.__model = get_model_by_name(self.__model)
        return self.__model

    @property
    def _manager(self):
        return self.__manager or self._model.objects

    def on_get(self, data):
        return self._manager.get(object_id=data)

    def on_set(self, value):
        if isinstance(value, self._model):
            return value.id
        return value


class _ListData:
    def __init__(self, data, instance, model, manager=None, links=LinksInteraction()):
        self._data = data
        self._model = model
        self._instance = instance
        self._manager = manager or model.objects
        self._links = links

    def __iter__(self):
        yield from (self._manager.get(item["id"]) for item in self._data)

    def append(self, value, main=False):
        return self._links.link(for_entity=self._instance, to_entity=value, main=main)

    add = append

    def remove(self, value):
        return self._links.unlink(for_entity=self._instance, to_entity=value)


class _EmbeddedLinkField(_Link):
    _path = [
        "_embedded",
    ]

    def on_get(self, data):
        if data:
            return self._model.objects.get(data[0]["id"])
        return None

    def on_set_instance(self, instance, value):
        if instance.id is None:
            raise exceptions.InitException("Create entity first")
        if value.id is None:
            value.create()
        self._links.link(instance, value)
        return [{"id": value.id}]


class _EmbeddedLinkListField(_Link):
    _path = [
        "_embedded",
    ]

    def on_get_instance(self, instance, value):
        return _ListData(data=value, model=self._model, manager=self._manager, instance=instance, links=self._links)

    def on_set(self, value):
        raise TypeError()



import pytest

from amocrm.v2 import Contact, exceptions

from .data.companies import DETAIL_INFO as COMPANY_DETAIL_INFO
from .data.contacts import CREATE_DATA, DETAIL_INFO, LIST_PAGE_1, UPDATE
from .data.users import DETAIL_INFO as USER_DETAIL_INFO


def test_get(response_mock):
    response_mock.add("GET", "https://test.amocrm.ru/api/v4/contacts/3", match_querystring=False, json=DETAIL_INFO)
    contact = Contact.objects.get(3)

    assert contact.id == 3
    assert contact.name == "Иван Иванов"
    assert contact.first_name == "Иван"
    assert contact.last_name == "Иванов"

    response_mock.add("GET", "https://test.amocrm.ru/api/v4/users/3", match_querystring=False, json=USER_DETAIL_INFO)
    assert contact.responsible_user.id == 3
    assert contact.responsible_user.name == "Алексей Поимцев"

    response_mock.add(
        "GET", "https://test.amocrm.ru/api/v4/companies/1", match_querystring=False, json=COMPANY_DETAIL_INFO
    )
    assert contact.company.name == "АО Рога и копыта"
    assert contact.company.account_id == 10


def test_get_not_found(response_mock):
    response_mock.add("GET", "https://test.amocrm.ru/api/v4/contacts/3", match_querystring=False, status=204)
    with pytest.raises(exceptions.NotFound):
        Contact.objects.get(3)


def test_list(response_mock):
    response_mock.add(
        "GET", "https://test.amocrm.ru/api/v4/contacts", match_querystring=False, status=200, json=LIST_PAGE_1
    )

    contacts = list(Contact.objects.all())
    assert len(contacts) == 2
    assert contacts[0].name == "1"


def test_create(response_mock):
    response_mock.add("POST", "https://test.amocrm.ru/api/v4/contacts", status=200, json=CREATE_DATA)
    c = Contact(name="test", responsible_user=3, tags=[])
    c.save()
    assert c.id == 3


def test_update(response_mock):
    response_mock.add("GET", "https://test.amocrm.ru/api/v4/contacts/3", match_querystring=False, json=DETAIL_INFO)
    response_mock.add("PATCH", "https://test.amocrm.ru/api/v4/contacts/3", match_querystring=False, json=UPDATE)

    contact = Contact.objects.get(3)
    contact.last_name = "e"
    contact.tags.add("tag")
    contact.save()



from .. import fields, manager, model
from ..interaction import GenericInteraction
from .note import NotesField
from .tag import TagsField


class CompaniesInteraction(GenericInteraction):
    path = "companies"


class Company(model.Model):
    name = fields._Field("name")
    responsible_user = fields._Link("responsible_user_id", "User")
    group = fields._Field("group_id")
    created_by = fields._Link("created_by", "User")
    updated_by = fields._Link("updated_by", "User")
    created_at = fields._DateTimeField("created_at")
    updated_at = fields._DateTimeField("updated_at")
    closest_task_at = fields._DateTimeField("closest_task_at", blank=True)
    account_id = fields._UnEditableField("account_id")

    tags = TagsField()
    notes = NotesField()

    leads = fields._EmbeddedLinkListField("leads", "Lead")
    customers = fields._EmbeddedLinkListField("customers", "Customer")
    contacts = fields._EmbeddedLinkListField("contacts", "Contact")

    objects = manager.Manager(CompaniesInteraction())



import logging
import os
from typing import Optional, Tuple

import jwt
import requests

from . import exceptions

logger = logging.getLogger(__name__)


class TokensStorage:
    def get_access_token(self) -> Optional[str]:
        pass

    def get_refresh_token(self) -> Optional[str]:
        pass

    def save_tokens(self, access_token: str, refresh_token: str):
        pass


class MemoryTokensStorage(TokensStorage):
    def __init__(self):
        self._access_token = None
        self._refresh_token = None

    def get_access_token(self) -> Optional[str]:
        return self._access_token

    def get_refresh_token(self) -> Optional[str]:
        return self._refresh_token

    def save_tokens(self, access_token: str, refresh_token: str):
        self._access_token, self._refresh_token = access_token, refresh_token


class FileTokensStorage(TokensStorage):
    def __init__(self, directory_path=os.getcwd()):
        self._access_token_path = os.path.join(directory_path, "access_token.txt")
        self._refresh_token_path = os.path.join(directory_path, "refresh_token.txt")

    @staticmethod
    def _read_file(path):
        try:
            with open(path, "r") as _file:
                return _file.read().strip()
        except FileNotFoundError:
            return None

    def get_access_token(self) -> Optional[str]:
        return self._read_file(self._access_token_path)

    def get_refresh_token(self) -> Optional[str]:
        return self._read_file(self._refresh_token_path)

    def save_tokens(self, access_token: str, refresh_token: str):
        with open(self._access_token_path, "w") as _file:
            _file.write(access_token)

        with open(self._refresh_token_path, "w") as _file:
            _file.write(refresh_token)


class RedisTokensStorage(TokensStorage):
    _ACCESS_TOKEN_KEY = "amocrm:access:token"
    _REFRESH_TOKEN_KEY = "amocrm:refresh:token"

    def __init__(self, client, ttl=None):
        self._ttl = ttl
        self._client = client

    def get_access_token(self) -> Optional[str]:
        token = self._client.get(self._ACCESS_TOKEN_KEY)
        if token:
            return token.decode()
        return None

    def get_refresh_token(self) -> Optional[str]:
        token = self._client.get(self._REFRESH_TOKEN_KEY)
        if token:
            return token.decode()
        return None

    def save_tokens(self, access_token: str, refresh_token: str):
        self._client.set(self._ACCESS_TOKEN_KEY, access_token, ex=self._ttl)
        self._client.set(self._REFRESH_TOKEN_KEY, refresh_token, ex=self._ttl)


class TokenManager:
    def __init__(self):
        self._client_id = None
        self._client_secret = None
        self.subdomain = None
        self._redirect_url = None
        self._storage: Optional[TokensStorage] = None

    def __call__(
        self, client_id: str, client_secret: str, subdomain: str, redirect_url: str, storage=FileTokensStorage()
    ):
        self._client_id = client_id
        self._client_secret = client_secret
        self._redirect_url = redirect_url
        if self._storage is None:
            self._storage = storage
        self.subdomain = subdomain

    def init(self, code, skip_error=False):
        data = {
            "grant_type": "authorization_code",
            "code": code,
            "redirect_uri": self._redirect_url,
            "client_id": self._client_id,
            "client_secret": self._client_secret,
        }
        try:
            response = requests.post("https://{}.amocrm.ru/oauth2/access_token".format(self.subdomain), json=data)
        except requests.exceptions.RequestException:
            logger.warning("can't init tokens")
            if not skip_error:
                raise
        else:
            if response.status_code != 200 and not skip_error:
                raise Exception(response.json()["title"])
            if response.status_code != 200 and skip_error:
                return
            response = response.json()
            self._storage.save_tokens(response["access_token"], response["refresh_token"])
            logger.info("succesful init and store tokens in %s store", self._storage)

    def _get_new_tokens(self) -> Tuple[str, str]:
        refresh_token = self._storage.get_refresh_token()
        if refresh_token is None:
            raise ValueError()
        body = {
            "client_id": self._client_id,
            "client_secret": self._client_secret,
            "grant_type": "refresh_token",
            "refresh_token": refresh_token,
            "redirect_uri": self._redirect_url,
        }
        response = requests.post("https://{}.amocrm.ru/oauth2/access_token".format(self.subdomain), json=body).json()
        return response["access_token"], response["refresh_token"]

    def get_access_token(self):
        token = self._storage.get_access_token()
        if token is None:
            raise exceptions.NoToken("You need to init tokens with code by 'init' method")
        if self._is_expire(token):
            token, refresh_token = self._get_new_tokens()
            self._storage.save_tokens(token, refresh_token)
        return token

    @staticmethod
    def _is_expire(token: str):
        alg = jwt.get_unverified_header(token).get("alg")
        try:
            jwt.decode(
                token,
                verify=True,
                options={"verify_exp": True, "verify_signature": False, "verify_aud": False},
                algorithms=[alg],
            )
        except jwt.exceptions.ExpiredSignatureError:
            return True
        else:
            return False


default_token_manager = TokenManager()



from .. import fields, manager, model
from ..interaction import GenericInteraction


class TasksInteraction(GenericInteraction):
    path = "tasks"


class Task(model.Model):

    name = fields._Field("name")
    created_by = fields._Link("created_by", "User")
    updated_by = fields._Link("updated_by", "User")
    created_at = fields._DateTimeField("created_at")
    updated_at = fields._DateTimeField("updated_at")
    responsible_user = fields._Link("responsible_user_id", "User")
    is_completed = fields._Field("is_completed")
    task_type_id = fields._Field("task_type_id")
    text = fields._Field("text")
    duration_sec = fields._Field("duration")
    complete_till = fields._DateTimeField("complete_till")
    result = fields._Field("text", path=["result"])

    objects = manager.Manager(TasksInteraction())



class Manager:
    def __init__(self, interaction, model=None):
        self._interaction = interaction
        self._model = model

    def __get__(self, instance, owner):
        if instance is None:
            self._model = owner
            return self
        raise AttributeError("Cant use with instance")

    def __set__(self, instance, value):
        raise TypeError()

    def create(self, *args, **kwargs):
        return self._model(data=self._interaction.create(*args, **kwargs))

    def update(self, *args, **kwargs):
        return self._interaction.update(*args, **kwargs)

    def get(self, object_id=None, query=None):
        if object_id is not None:
            return self._model(data=self._interaction.get(object_id, include=self._model._get_embedded_fields()))
        return next(self.filter(query=query))

    def filter(self, *args, **kwargs):
        for data in self._interaction.get_all(*args, include=self._model._get_embedded_fields(), **kwargs):
            yield self._model(data=data)

    def all(self):
        return self.filter()



class _RegisterMeta(type):
    _REGISTER = {}

    def __new__(cls, name, bases, dct):
        _class = super().__new__(cls, name, bases, dct)
        cls._REGISTER[name] = _class
        return _class

    @classmethod
    def get_model_by_name(cls, name):
        return cls._REGISTER.get(name)


def get_model_by_name(name):
    return _RegisterMeta.get_model_by_name(name)



from .. import fields, manager, model
from ..interaction import GenericInteraction


class Tag(model.Model):
    name = fields._Field("name")

    leads = manager.Manager(GenericInteraction(path="leads/tags"))
    contacts = manager.Manager(GenericInteraction(path="contacts/tags"))
    companies = manager.Manager(GenericInteraction(path="companies/tags"))
    customers = manager.Manager(GenericInteraction(path="customers/tags"))


class _TagsList:
    def __init__(self, tags, on_change):
        """
        BAD CODE THERE
        self._tags is mutable object and this class and arch play with it
        """
        self._tags = tags
        self._on_change = on_change

    def __iter__(self):
        yield from (Tag(data=item) for item in self._tags)

    def _add(self, tag):
        self._tags.append(tag)
        self._on_change()

    def append(self, value):
        if isinstance(value, Tag):
            if value.id is None:
                value.create()
            self._add({"id": value.id})
        else:
            self._add({"name": value})

    add = append

    def remove(self, value):
        if isinstance(value, Tag):
            value = value.id
        self._remove(value)

    def _remove(self, value):
        for tag in self._tags:
            if tag["id"] == value or tag["name"] == value:
                self._tags.remove(tag)
                self._on_change()
                return


class TagsField(fields._BaseField):
    _path = ["_embedded"]

    def __init__(self):
        super().__init__("tags", blank=True)

    def on_get_instance(self, instance, tags):
        if tags is None:
            tags = instance._data.setdefault("_embedded", {}).setdefault("tags", [])
        return _TagsList(tags, on_change=lambda: self._notify_instance(instance))



#!/usr/bin/env python
from setuptools import setup, find_packages
import re


def get_version():
    init_py = open('amocrm/__init__.py').read()
    metadata = dict(re.findall("__([a-z]+)__ = '([^']+)'", init_py))
    return metadata['version']

version = get_version()


setup(
    name='amocrm_api',
    version=version,
    packages=find_packages(),
    url='https://github.com/Krukov/amocrm_api',
    download_url='https://github.com/Krukov/amocrm_api/tarball/%s' % version,
    license='MIT license',
    author='Dmitry Kryukov',
    author_email='glebov.ru@gmail.com',
    description='Python API for Amocrm',
    long_description=open('README.rst').read(),
    requires=[
        'requests',
        'pyjwt'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

__version__ = '2.0.1'

import sys
import logging

from .settings import settings as amo_settings
from .api import *
from .apimodels import *
from . import fields

__all__ = [
    'BaseCompany', 'BaseContact', 'BaseLead', 'amo_settings', 'AmoApi', 'ContactNote', 'ContactTask',
    'LeadNote', 'LeadTask', 'fields',
]

logger = logging.getLogger('amocrm')

if not logger.handlers:
    formatter = logging.Formatter(
        '%(asctime)s.%(msecs)d %(levelname)s: %(message)s', '%Y-%m-%d %H:%M:%S'
    )
    handler = logging.StreamHandler(sys.stderr)
    handler.setFormatter(formatter)
    handler.setLevel(logging.DEBUG)
    logger.addHandler(handler)



DETAIL_INFO = {
    "id": 3,
    "name": "Иван Иванов",
    "first_name": "Иван",
    "last_name": "Иванов",
    "responsible_user_id": 3,
    "group_id": 0,
    "created_by": 3,
    "updated_by": 504141,
    "created_at": 1582117331,
    "updated_at": 1590943929,
    "closest_task_at": None,
    "custom_fields_values": [
        {
            "field_id": 3,
            "field_name": "Телефон",
            "field_code": "PHONE",
            "field_type": "multitext",
            "values": [{"value": "+79123", "enum_id": 1, "enum": "WORK"}],
        }
    ],
    "account_id": 28805383,
    "_links": {"self": {"href": "https://example.amocrm.ru/api/v4/contacts/3"}},
    "_embedded": {
        "tags": [],
        "leads": [
            {"id": 1, "_links": {"self": {"href": "https://example.amocrm.ru/api/v4/leads/1"}}},
            {"id": 3916883, "_links": {"self": {"href": "https://example.amocrm.ru/api/v4/leads/3916883"}}},
        ],
        "customers": [
            {"id": 134923, "_links": {"self": {"href": "https://example.amocrm.ru/api/v4/customers/134923"}}}
        ],
        "catalog_elements": [],
        "companies": [{"id": 1, "_links": {"self": {"href": "https://example.amocrm.ru/api/v4/companies/1"}}}],
    },
}

LIST_PAGE_1 = {
    "_page": 1,
    "_links": {
        "self": {"href": "https://example.amocrm.ru/api/v4/contacts?limit=2&page=1"},
        "next": {"href": "https://example.amocrm.ru/api/v4/contacts?limit=2&page=2"},
    },
    "_embedded": {
        "contacts": [
            {
                "id": 7143599,
                "name": "1",
                "first_name": "",
                "last_name": "",
                "responsible_user_id": 504141,
                "group_id": 0,
                "created_by": 504141,
                "updated_by": 504141,
                "created_at": 1585758065,
                "updated_at": 1585758065,
                "closest_task_at": None,
                "custom_fields_values": None,
                "account_id": 28805383,
                "_links": {"self": {"href": "https://example.amocrm.ru/api/v4/contacts/7143599"}},
                "_embedded": {"tags": [], "companies": []},
            },
            {
                "id": 7767065,
                "name": "dsgdsg",
                "first_name": "",
                "last_name": "",
                "responsible_user_id": 504141,
                "group_id": 0,
                "created_by": 504141,
                "updated_by": 504141,
                "created_at": 1586359590,
                "updated_at": 1586359590,
                "closest_task_at": None,
                "custom_fields_values": None,
                "account_id": 28805383,
                "_links": {"self": {"href": "https://example.amocrm.ru/api/v4/contacts/7767065"}},
                "_embedded": {"tags": [], "companies": []},
            },
        ]
    },
}

CREATE_DATA = {
    "_links": {"self": {"href": "https://example.amocrm.ru/api/v4/contacts"}},
    "_embedded": {
        "contacts": [
            {
                "id": 3,
                "request_id": "0",
                "_links": {"self": {"href": "https://example.amocrm.ru/api/v4/contacts/40401635"}},
            },
        ]
    },
}

UPDATE = {
    "_links": {"self": {"href": "https://example.amocrm.ru/api/v4/contacts"}},
    "_embedded": {
        "contacts": [
            {
                "id": 3,
                "name": "Иван Иванов",
                "updated_at": 1590945248,
                "_links": {"self": {"href": "https://example.amocrm.ru/api/v4/contacts/3"}},
            }
        ]
    },
}
